(* Code by Leonid Shifrin https://stackoverflow.com/a/8746584/2576097 *)

BeginPackage["AtomicDensityMatrix`SelectiveEvaluate`"];

SelectiveEvaluate::usage="";

Begin["`Private`"];

SetAttributes[symbolToHideQ, HoldFirst];
symbolToHideQ[s_Symbol, expandedSymbs_] := ! MemberQ[expandedSymbs, Unevaluated[s]];

globalProperties[] := {DownValues, SubValues, UpValues (*,OwnValues*)};

Options[getSymbolsToHide] = {
    Exceptions -> {List, Hold, HoldComplete,
        HoldForm, HoldPattern, Blank, BlankSequence, BlankNullSequence,
        Optional, Repeated, Verbatim, Pattern, RuleDelayed, Rule, True,
        False, Integer, Real, Complex, Alternatives, String,
        PatternTest, (* Note-this one is dangerous since it opens a hole to evaluation leaks. But too good to be ignored *)
        Condition, PatternSequence, Except
    }
};

getSymbolsToHide[code_Hold, headsToExpand : {___Symbol}, opts : OptionsPattern[]] :=
    Join @@ Complement[
        Cases[{
            Flatten[Outer[Compose, globalProperties[], headsToExpand]], code},
            s_Symbol /; symbolToHideQ[s, headsToExpand] :> Hold[s],
            Infinity,
            Heads -> True
        ],
        Hold /@ OptionValue[Exceptions]
    ];

SetAttributes[makeHidingSymbol, HoldAll];
makeHidingSymbol[s_Symbol] :=
    Unique[hidingSymbol(*Unevaluated[s]*) (*,Attributes[s]*)];

makeHidingRules[symbs : Hold[__Symbol]] :=
    Thread[List @@ Map[HoldPattern, symbs] -> List @@ Map[makeHidingSymbol, symbs]];

reverseHidingRules[rules : {(_Rule | _RuleDelayed) ..}] :=
    rules /. (Rule | RuleDelayed)[Verbatim[HoldPattern][lhs_], rhs_] :> (rhs :> lhs);

SelectiveEvaluate[code_Hold, headsToEvaluate : {___Symbol}] :=
    Module[{symbolsToHide, hidingRules, revHidingRules, result},
        symbolsToHide = getSymbolsToHide[code, headsToEvaluate];
        hidingRules = makeHidingRules@symbolsToHide;
        revHidingRules = reverseHidingRules@hidingRules;
        result = Hold@Evaluate@ReleaseHold[code /. hidingRules] /. revHidingRules;
        Apply[Remove, revHidingRules[[All, 1]]];
        result
    ];

End[];

EndPackage[];